@@ -15,6 +15,7 @@
 #include <stdlib.h>
 #include <math.h>
 #include "gmem.h"
+#include "gmempp.h"
 #include "GString.h"
 #include "GList.h"
 #include "Error.h"
@@ -79,33 +80,105 @@ public:
 //------------------------------------------------------------------------
 
 AcroForm *AcroForm::load(PDFDoc *docA, Catalog *catalog, Object *acroFormObjA) {
+  Object acroFormObj2;
   AcroForm *acroForm;
-  Object fieldsObj, obj1, obj2;
-  int i;
+  AcroFormField *field;
+  Object fieldsObj, annotsObj, annotRef, annotObj, obj1, obj2;
+  int pageNum, i, j;
 
-  acroForm = new AcroForm(docA, acroFormObjA);
+  // this is the normal case: acroFormObj is a dictionary, as expected
+  if (acroFormObjA->isDict()) {
+    acroForm = new AcroForm(docA, acroFormObjA);
 
-  if (acroFormObjA->dictLookup("NeedAppearances", &obj1)->isBool()) {
-    acroForm->needAppearances = obj1.getBool();
-  }
-  obj1.free();
-
-  acroForm->buildAnnotPageList(catalog);
-
-  if (!acroFormObjA->dictLookup("Fields", &obj1)->isArray()) {
-    if (!obj1.isNull()) {
-      error(errSyntaxError, -1, "AcroForm Fields entry is wrong type");
+    if (acroFormObjA->dictLookup("NeedAppearances", &obj1)->isBool()) {
+      acroForm->needAppearances = obj1.getBool();
     }
     obj1.free();
-    delete acroForm;
-    return NULL;
+
+    acroForm->buildAnnotPageList(catalog);
+
+    if (!acroFormObjA->dictLookup("Fields", &obj1)->isArray()) {
+      if (!obj1.isNull()) {
+	error(errSyntaxError, -1, "AcroForm Fields entry is wrong type");
+      }
+      obj1.free();
+      delete acroForm;
+      return NULL;
+    }
+    for (i = 0; i < obj1.arrayGetLength(); ++i) {
+      obj1.arrayGetNF(i, &obj2);
+      acroForm->scanField(&obj2);
+      obj2.free();
+    }
+    obj1.free();
+
+    // scan the annotations, looking for Widget-type annots that are
+    // not attached to the AcroForm object
+    for (pageNum = 1; pageNum <= catalog->getNumPages(); ++pageNum) {
+      if (catalog->getPage(pageNum)->getAnnots(&annotsObj)->isArray()) {
+	for (i = 0; i < annotsObj.arrayGetLength(); ++i) {
+	  if (annotsObj.arrayGetNF(i, &annotRef)->isRef()) {
+	    for (j = 0; j < acroForm->fields->getLength(); ++j) {
+	      field = (AcroFormField *)acroForm->fields->get(j);
+	      if (field->fieldRef.isRef()) {
+		if (field->fieldRef.getRefNum() == annotRef.getRefNum() &&
+		    field->fieldRef.getRefGen() == annotRef.getRefGen()) {
+		  break;
+		}
+	      }
+	    }
+	    if (j == acroForm->fields->getLength()) {
+	      annotRef.fetch(acroForm->doc->getXRef(), &annotObj);
+	      if (annotObj.isDict()) {
+		if (annotObj.dictLookup("Subtype", &obj1)->isName("Widget")) {
+		  acroForm->scanField(&annotRef);
+		}
+		obj1.free();
+	      }
+	      annotObj.free();
+	    }
+	  }
+	  annotRef.free();
+	}
+      }
+      annotsObj.free();
+    }
+
+  // if acroFormObjA is a null object, but there are Widget-type
+  // annots, we still create an AcroForm
+  } else {
+    // create an empty dict for acroFormObj
+    acroFormObj2.initDict(docA->getXRef());
+    acroForm = new AcroForm(docA, &acroFormObj2);
+    acroFormObj2.free();
+
+    acroForm->buildAnnotPageList(catalog);
+
+    // scan the annotations, looking for any Widget-type annots
+    for (pageNum = 1; pageNum <= catalog->getNumPages(); ++pageNum) {
+      if (catalog->getPage(pageNum)->getAnnots(&annotsObj)->isArray()) {
+	for (i = 0; i < annotsObj.arrayGetLength(); ++i) {
+	  if (annotsObj.arrayGetNF(i, &annotRef)->isRef()) {
+	    annotRef.fetch(acroForm->doc->getXRef(), &annotObj);
+	    if (annotObj.isDict()) {
+	      if (annotObj.dictLookup("Subtype", &obj1)->isName("Widget")) {
+		acroForm->scanField(&annotRef);
+	      }
+	      obj1.free();
+	    }
+	    annotObj.free();
+	  }
+	  annotRef.free();
+	}
+      }
+      annotsObj.free();
+    }
+
+    if (acroForm->fields->getLength() == 0) {
+      delete acroForm;
+      acroForm = NULL;
+    }
   }
-  for (i = 0; i < obj1.arrayGetLength(); ++i) {
-    obj1.arrayGetNF(i, &obj2);
-    acroForm->scanField(&obj2);
-    obj2.free();
-  }
-  obj1.free();
 
   return acroForm;
 }
@@ -235,7 +308,7 @@ AcroFormField *AcroFormField::load(AcroForm *acroFormA, Object *fieldRefA) {
   GString *typeStr;
   TextString *nameA;
   Guint flagsA;
-  GBool haveFlags;
+  GBool haveFlags, typeFromParentA;
   Object fieldObjA, parentObj, parentObj2, obj1, obj2;
   AcroFormFieldType typeA;
   AcroFormField *field;
@@ -253,8 +326,10 @@ AcroFormField *AcroFormField::load(AcroForm *acroFormA, Object *fieldRefA) {
 
   if (fieldObjA.dictLookup("FT", &obj1)->isName()) {
     typeStr = new GString(obj1.getName());
+    typeFromParentA = gFalse;
   } else {
     typeStr = NULL;
+    typeFromParentA = gTrue;
   }
   obj1.free();
 
@@ -304,7 +379,19 @@ AcroFormField *AcroFormField::load(AcroForm *acroFormA, Object *fieldRefA) {
   if (!typeStr) {
     error(errSyntaxError, -1, "Missing type in AcroForm field");
     goto err1;
-  } else if (!typeStr->cmp("Btn")) {
+  }
+
+  //----- check for a radio button
+
+  // this is a kludge: if we see a Btn-type field with kids, and the
+  // Ff entry is missing, assume the kids are radio buttons
+  if (typeFromParentA && !typeStr->cmp("Btn") && !haveFlags) {
+    flagsA = acroFormFlagRadio;
+  }
+
+  //----- determine field type
+
+  if (!typeStr->cmp("Btn")) {
     if (flagsA & acroFormFlagPushbutton) {
       typeA = acroFormFieldPushbutton;
     } else if (flagsA & acroFormFlagRadio) {
@@ -335,7 +422,7 @@ AcroFormField *AcroFormField::load(AcroForm *acroFormA, Object *fieldRefA) {
   delete typeStr;
 
   field = new AcroFormField(acroFormA, fieldRefA, &fieldObjA,
-			    typeA, nameA, flagsA);
+			    typeA, nameA, flagsA, typeFromParentA);
   fieldObjA.free();
   return field;
 
@@ -349,13 +436,14 @@ AcroFormField *AcroFormField::load(AcroForm *acroFormA, Object *fieldRefA) {
 AcroFormField::AcroFormField(AcroForm *acroFormA,
 			     Object *fieldRefA, Object *fieldObjA,
 			     AcroFormFieldType typeA, TextString *nameA,
-			     Guint flagsA) {
+			     Guint flagsA, GBool typeFromParentA) {
   acroForm = acroFormA;
   fieldRefA->copy(&fieldRef);
   fieldObjA->copy(&fieldObj);
   type = typeA;
   name = nameA;
   flags = flagsA;
+  typeFromParent = typeFromParentA;
 }
 
 AcroFormField::~AcroFormField() {
@@ -364,6 +452,24 @@ AcroFormField::~AcroFormField() {
   delete name;
 }
 
+int AcroFormField::getPageNum() {
+  Object kidsObj, annotRef;
+  int pageNum;
+
+  pageNum = 0;
+  if (fieldObj.dictLookup("Kids", &kidsObj)->isArray()) {
+    if (kidsObj.arrayGetLength() > 0) {
+      kidsObj.arrayGetNF(0, &annotRef);
+      pageNum = acroForm->lookupAnnotPage(&annotRef);
+      annotRef.free();
+    }
+  } else {
+    pageNum = acroForm->lookupAnnotPage(&fieldRef);
+  }
+  kidsObj.free();
+  return pageNum;
+}
+
 const char *AcroFormField::getType() {
   switch (type) {
   case acroFormFieldPushbutton:    return "PushButton";
@@ -392,12 +498,16 @@ Unicode *AcroFormField::getName(int *length) {
 }
 
 Unicode *AcroFormField::getValue(int *length) {
-  Object obj1;
+  Object obj1, obj2;
   Unicode *u;
   char *s;
   TextString *ts;
+  GString *gs;
   int n, i;
 
+  u = NULL;
+  *length = 0;
+
   fieldLookup("V", &obj1);
   if (obj1.isName()) {
     s = obj1.getName();
@@ -407,7 +517,6 @@ Unicode *AcroFormField::getValue(int *length) {
       u[i] = s[i] & 0xff;
     }
     *length = n;
-    return u;
   } else if (obj1.isString()) {
     ts = new TextString(obj1.getString());
     n = ts->getLength();
@@ -415,10 +524,198 @@ Unicode *AcroFormField::getValue(int *length) {
     memcpy(u, ts->getUnicode(), n * sizeof(Unicode));
     *length = n;
     delete ts;
-    return u;
-  } else {
-    return NULL;
+  } else if (obj1.isDict()) {
+    obj1.dictLookup("Contents", &obj2);
+    if (obj2.isString()) {
+      gs = obj2.getString();
+      n = gs->getLength();
+      u = (Unicode *)gmallocn(n, sizeof(Unicode));
+      for (i = 0; i < n; ++i) {
+	u[i] = gs->getChar(i) & 0xff;
+      }
+      *length = n;
+    }
+    obj2.free();
   }
+  obj1.free();
+
+  return u;
+}
+
+void AcroFormField::getBBox(double *llx, double *lly,
+			    double *urx, double *ury) {
+  Object annotObj, rectObj, numObj;
+  double t;
+
+  *llx = *lly = *urx = *ury = 0;
+
+  if (getAnnotObj(&annotObj)->isDict()) {
+    annotObj.dictLookup("Rect", &rectObj);
+    if (rectObj.isArray() && rectObj.arrayGetLength() == 4) {
+      rectObj.arrayGet(0, &numObj);
+      if (numObj.isNum()) {
+	*llx = numObj.getNum();
+      }
+      numObj.free();
+      rectObj.arrayGet(1, &numObj);
+      if (numObj.isNum()) {
+	*lly = numObj.getNum();
+      }
+      numObj.free();
+      rectObj.arrayGet(2, &numObj);
+      if (numObj.isNum()) {
+	*urx = numObj.getNum();
+      }
+      numObj.free();
+      rectObj.arrayGet(3, &numObj);
+      if (numObj.isNum()) {
+	*ury = numObj.getNum();
+      }
+      numObj.free();
+    }
+    rectObj.free();
+  }
+  annotObj.free();
+
+  if (*llx > *urx) {
+    t = *llx;  *llx = *urx;  *urx = t;
+  }
+  if (*lly > *ury) {
+    t = *lly;  *lly = *ury;  *ury = t;
+  }
+}
+
+void AcroFormField::getFont(Ref *fontID, double *fontSize) {
+  Object daObj;
+  GList *daToks;
+  char *fontTag;
+  double tfSize, m2, m3;
+  int tfPos, tmPos, i;
+
+  fontID->num = fontID->gen = -1;
+  *fontSize = 0;
+
+  if (fieldLookup("DA", &daObj)->isString()) {
+
+    // parse the default appearance string
+    daToks = tokenize(daObj.getString());
+    tfPos = tmPos = -1;
+    for (i = 2; i < daToks->getLength(); ++i) {
+      if (!((GString *)daToks->get(i))->cmp("Tf")) {
+	tfPos = i - 2;
+      } else if (i >= 6 && !((GString *)daToks->get(i))->cmp("Tm")) {
+	tmPos = i - 6;
+      }
+    }
+
+    // handle the Tf operator
+    if (tfPos >= 0) {
+      fontTag = ((GString *)daToks->get(tfPos))->getCString();
+      if (*fontTag == '/') {
+	++fontTag;
+      }
+      *fontID = findFontName(fontTag);
+      tfSize = atof(((GString *)daToks->get(tfPos + 1))->getCString());
+    } else {
+      tfSize = 1;
+    }
+
+    // handle the Tm operator
+    if (tmPos >= 0) {
+      // transformed font size = sqrt(m[2]^2 + m[3]^2) * size
+      m2 = atof(((GString *)daToks->get(tfPos + 2))->getCString());
+      m3 = atof(((GString *)daToks->get(tfPos + 3))->getCString());
+      tfSize *= sqrt(m2*m2 + m3*m3);
+    }
+    *fontSize = tfSize;
+
+    deleteGList(daToks, GString);
+  }
+
+  daObj.free();
+}
+
+Ref AcroFormField::findFontName(char *fontTag) {
+  Object drObj, fontDictObj, fontObj, baseFontObj;
+  Ref fontID;
+  GBool found;
+
+  fontID.num = fontID.gen = -1;
+  found = gFalse;
+
+  if (fieldObj.dictLookup("DR", &drObj)->isDict()) {
+    if (drObj.dictLookup("Font", &fontDictObj)->isDict()) {
+      if (fontDictObj.dictLookupNF(fontTag, &fontObj)->isRef()) {
+	fontID = fontObj.getRef();
+	found = gTrue;
+      }
+      fontObj.free();
+    }
+    fontDictObj.free();
+  }
+  drObj.free();
+  if (found) {
+    return fontID;
+  }
+
+  if (acroForm->acroFormObj.dictLookup("DR", &drObj)->isDict()) {
+    if (drObj.dictLookup("Font", &fontDictObj)->isDict()) {
+      if (fontDictObj.dictLookupNF(fontTag, &fontObj)->isRef()) {
+	fontID = fontObj.getRef();
+      }
+      fontObj.free();
+    }
+    fontDictObj.free();
+  }
+  drObj.free();
+  return fontID;
+}
+
+void AcroFormField::getColor(double *red, double *green, double *blue) {
+  Object daObj;
+  GList *daToks;
+  int i;
+
+  *red = *green = *blue = 0;
+
+  if (fieldLookup("DA", &daObj)->isString()) {
+
+    // parse the default appearance string
+    daToks = tokenize(daObj.getString());
+    for (i = 1; i < daToks->getLength(); ++i) {
+
+      // handle the g operator
+      if (!((GString *)daToks->get(i))->cmp("g")) {
+	*red = *green = *blue =
+	    atof(((GString *)daToks->get(i - 1))->getCString());
+	break;
+
+      // handle the rg operator
+      } else if (i >= 3 && !((GString *)daToks->get(i))->cmp("rg")) {
+	*red = atof(((GString *)daToks->get(i - 3))->getCString());
+	*green = atof(((GString *)daToks->get(i - 2))->getCString());
+	*blue = atof(((GString *)daToks->get(i - 1))->getCString());
+	break;
+      }
+    }
+
+    deleteGList(daToks, GString);
+  }
+
+  daObj.free();
+}
+
+int AcroFormField::getMaxLen() {
+  Object obj;
+  int len;
+
+  if (fieldLookup("MaxLen", &obj)->isInt()) {
+    len = obj.getInt();
+  } else {
+    len = -1;
+  }
+  obj.free();
+  return len;
 }
 
 void AcroFormField::draw(int pageNum, Gfx *gfx, GBool printing) {
@@ -445,7 +742,7 @@ void AcroFormField::drawAnnot(int pageNum, Gfx *gfx, GBool printing,
   Object obj1, obj2;
   double xMin, yMin, xMax, yMax, t;
   int annotFlags;
-  GBool oc;
+  GBool oc, render;
 
   if (!annotObj->isDict()) {
     return;
@@ -518,7 +815,16 @@ void AcroFormField::drawAnnot(int pageNum, Gfx *gfx, GBool printing,
 
   //----- draw it
 
+  render = gFalse;
   if (acroForm->needAppearances) {
+    render = gTrue;
+  } else {
+    if (!annotObj->dictLookup("AP", &obj1)->isDict()) {
+      render = gTrue;
+    }
+    obj1.free();
+  }
+  if (render) {
     drawNewAppearance(gfx, annotObj->getDict(),
 		      xMin, yMin, xMax, yMax);
   } else {
@@ -562,19 +868,22 @@ void AcroFormField::drawExistingAppearance(Gfx *gfx, Dict *annot,
   }
 }
 
-// Regenerate the appearnce for this field, and draw it.
+// Regenerate the appearance for this field, and draw it.
 void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
 				      double xMin, double yMin,
 				      double xMax, double yMax) {
+  GString *appearBuf;
   Object appearance, mkObj, ftObj, appearDict, drObj, apObj, asObj;
-  Object obj1, obj2, obj3;
+  Object resources, fontResources, defaultFont, gfxStateDict;
+  Object obj1, obj2, obj3, obj4;
   Dict *mkDict;
   MemStream *appearStream;
   GfxFontDict *fontDict;
   GBool hasCaption;
   double dx, dy, r;
-  GString *caption, *da;
+  GString *val, *caption, *da;
   GString **text;
+  GBool done;
   GBool *selection;
   AnnotBorderType borderType;
   double borderWidth;
@@ -595,7 +904,7 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
   if (mkDict) {
     if (mkDict->lookup("BG", &obj1)->isArray() &&
 	obj1.arrayGetLength() > 0) {
-      setColor(obj1.getArray(), gTrue, 0);
+      setColor(obj1.getArray(), gTrue, 0, appearBuf);
       appearBuf->appendf("0 0 {0:.4f} {1:.4f} re f\n",
 			 xMax - xMin, yMax - yMin);
     }
@@ -678,6 +987,7 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
     if (borderWidth > 0) {
       mkDict->lookup("BC", &obj1);
       if (!(obj1.isArray() && obj1.arrayGetLength() > 0)) {
+	obj1.free();
 	mkDict->lookup("BG", &obj1);
       }
       if (obj1.isArray() && obj1.arrayGetLength() > 0) {
@@ -700,20 +1010,26 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
 	  case annotBorderSolid:
 	  case annotBorderUnderlined:
 	    appearBuf->appendf("{0:.4f} w\n", borderWidth);
-	    setColor(obj1.getArray(), gFalse, 0);
-	    drawCircle(0.5 * dx, 0.5 * dy, r - 0.5 * borderWidth, "s");
+	    setColor(obj1.getArray(), gFalse, 0, appearBuf);
+	    drawCircle(0.5 * dx, 0.5 * dy, r - 0.5 * borderWidth, "s",
+		       appearBuf);
 	    break;
 	  case annotBorderBeveled:
 	  case annotBorderInset:
 	    appearBuf->appendf("{0:.4f} w\n", 0.5 * borderWidth);
-	    setColor(obj1.getArray(), gFalse, 0);
-	    drawCircle(0.5 * dx, 0.5 * dy, r - 0.25 * borderWidth, "s");
+	    setColor(obj1.getArray(), gFalse, 0, appearBuf);
+	    drawCircle(0.5 * dx, 0.5 * dy, r - 0.25 * borderWidth, "s",
+		       appearBuf);
 	    setColor(obj1.getArray(), gFalse,
-		     borderType == annotBorderBeveled ? 1 : -1);
-	    drawCircleTopLeft(0.5 * dx, 0.5 * dy, r - 0.75 * borderWidth);
+		     borderType == annotBorderBeveled ? 1 : -1,
+		     appearBuf);
+	    drawCircleTopLeft(0.5 * dx, 0.5 * dy, r - 0.75 * borderWidth,
+			      appearBuf);
 	    setColor(obj1.getArray(), gFalse,
-		     borderType == annotBorderBeveled ? -1 : 1);
-	    drawCircleBottomRight(0.5 * dx, 0.5 * dy, r - 0.75 * borderWidth);
+		     borderType == annotBorderBeveled ? -1 : 1,
+		     appearBuf);
+	    drawCircleBottomRight(0.5 * dx, 0.5 * dy, r - 0.75 * borderWidth,
+				  appearBuf);
 	    break;
 	  }
 
@@ -728,7 +1044,7 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
 	    // fall through to the solid case
 	  case annotBorderSolid:
 	    appearBuf->appendf("{0:.4f} w\n", borderWidth);
-	    setColor(obj1.getArray(), gFalse, 0);
+	    setColor(obj1.getArray(), gFalse, 0, appearBuf);
 	    appearBuf->appendf("{0:.4f} {0:.4f} {1:.4f} {2:.4f} re s\n",
 			       0.5 * borderWidth,
 			       dx - borderWidth, dy - borderWidth);
@@ -736,7 +1052,8 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
 	  case annotBorderBeveled:
 	  case annotBorderInset:
 	    setColor(obj1.getArray(), gTrue,
-		     borderType == annotBorderBeveled ? 1 : -1);
+		     borderType == annotBorderBeveled ? 1 : -1,
+		     appearBuf);
 	    appearBuf->append("0 0 m\n");
 	    appearBuf->appendf("0 {0:.4f} l\n", dy);
 	    appearBuf->appendf("{0:.4f} {1:.4f} l\n", dx, dy);
@@ -747,7 +1064,8 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
 	    appearBuf->appendf("{0:.4f} {0:.4f} l\n", borderWidth);
 	    appearBuf->append("f\n");
 	    setColor(obj1.getArray(), gTrue,
-		     borderType == annotBorderBeveled ? -1 : 1);
+		     borderType == annotBorderBeveled ? -1 : 1,
+		     appearBuf);
 	    appearBuf->append("0 0 m\n");
 	    appearBuf->appendf("{0:.4f} 0 l\n", dx);
 	    appearBuf->appendf("{0:.4f} {1:.4f} l\n", dx, dy);
@@ -760,7 +1078,7 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
 	    break;
 	  case annotBorderUnderlined:
 	    appearBuf->appendf("{0:.4f} w\n", borderWidth);
-	    setColor(obj1.getArray(), gFalse, 0);
+	    setColor(obj1.getArray(), gFalse, 0, appearBuf);
 	    appearBuf->appendf("0 0 m {0:.4f} 0 l s\n", dx);
 	    break;
 	  }
@@ -777,7 +1095,7 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
   gfree(borderDash);
 
   // get the resource dictionary
-  fieldLookup("DR", &drObj);
+  buildDefaultResourceDict(&drObj);
 
   // build the font dictionary
   if (drObj.isDict() && drObj.dictLookup("Font", &obj1)->isDict()) {
@@ -839,15 +1157,17 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
 	    ->isName(appearanceState->getCString())) {
 	if (caption) {
 	  drawText(caption, da, fontDict, gFalse, 0, acroFormQuadCenter,
-		   gFalse, gTrue, rot, xMin, yMin, xMax, yMax, borderWidth);
+		   gFalse, gTrue, rot, xMin, yMin, xMax, yMax, borderWidth,
+		   appearBuf);
 	} else {
 	  if (mkDict) {
 	    if (mkDict->lookup("BC", &obj2)->isArray() &&
 		obj2.arrayGetLength() > 0) {
 	      dx = xMax - xMin;
 	      dy = yMax - yMin;
-	      setColor(obj2.getArray(), gTrue, 0);
-	      drawCircle(0.5 * dx, 0.5 * dy, 0.2 * (dx < dy ? dx : dy), "f");
+	      setColor(obj2.getArray(), gTrue, 0, appearBuf);
+	      drawCircle(0.5 * dx, 0.5 * dy, 0.2 * (dx < dy ? dx : dy), "f",
+			 appearBuf);
 	    }
 	    obj2.free();
 	  }
@@ -858,17 +1178,21 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
     } else if (flags & acroFormFlagPushbutton) {
       if (caption) {
 	drawText(caption, da, fontDict, gFalse, 0, acroFormQuadCenter,
-		 gFalse, gFalse, rot, xMin, yMin, xMax, yMax, borderWidth);
+		 gFalse, gFalse, rot, xMin, yMin, xMax, yMax, borderWidth,
+		 appearBuf);
       }
     // checkbox
     } else {
       fieldLookup("V", &obj1);
-      if (obj1.isName() && !obj1.isName("Off")) {
+      if (obj1.isName() && !(obj1.isName("Off") ||
+			     obj1.isName("No") ||
+			     obj1.isName(""))) {
 	if (!caption) {
 	  caption = new GString("3"); // ZapfDingbats checkmark
 	}
 	drawText(caption, da, fontDict, gFalse, 0, acroFormQuadCenter,
-		 gFalse, gTrue, rot, xMin, yMin, xMax, yMax, borderWidth);
+		 gFalse, gTrue, rot, xMin, yMin, xMax, yMax, borderWidth,
+		 appearBuf);
       }
       obj1.free();
     }
@@ -877,10 +1201,7 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
     }
   } else if (ftObj.isName("Tx")) {
     //~ value strings can be Unicode
-    if (!fieldLookup("V", &obj1)->isString()) {
-      obj1.free();
-      fieldLookup("DV", &obj1);
-    }
+    fieldLookup("V", &obj1);
     if (obj1.isString()) {
       if (fieldLookup("Q", &obj2)->isInt()) {
 	quadding = obj2.getInt();
@@ -897,7 +1218,8 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
       }
       drawText(obj1.getString(), da, fontDict,
 	       flags & acroFormFlagMultiline, comb, quadding,
-	       gTrue, gFalse, rot, xMin, yMin, xMax, yMax, borderWidth);
+	       gTrue, gFalse, rot, xMin, yMin, xMax, yMax, borderWidth,
+	       appearBuf);
     }
     obj1.free();
   } else if (ftObj.isName("Ch")) {
@@ -911,9 +1233,30 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
     // combo box
     if (flags & acroFormFlagCombo) {
       if (fieldLookup("V", &obj1)->isString()) {
-	drawText(obj1.getString(), da, fontDict,
+	val = obj1.getString()->copy();
+	if (fieldObj.dictLookup("Opt", &obj2)->isArray()) {
+	  for (i = 0, done = false; i < obj2.arrayGetLength() && !done; ++i) {
+	    obj2.arrayGet(i, &obj3);
+	    if (obj3.isArray() && obj3.arrayGetLength() == 2) {
+	      if (obj3.arrayGet(0, &obj4)->isString() &&
+		  obj4.getString()->cmp(val) == 0) {
+		obj4.free();
+		if (obj3.arrayGet(1, &obj4)->isString()) {
+		  delete val;
+		  val = obj4.getString()->copy();
+		}
+		done = gTrue;
+	      }
+	      obj4.free();
+	    }
+	    obj3.free();
+	  }
+	}
+	obj2.free();
+	drawText(val, da, fontDict,
 		 gFalse, 0, quadding, gTrue, gFalse, rot,
-		 xMin, yMin, xMax, yMax, borderWidth);
+		 xMin, yMin, xMax, yMax, borderWidth, appearBuf);
+	delete val;
 	//~ Acrobat draws a popup icon on the right side
       }
       obj1.free();
@@ -969,7 +1312,7 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
 	obj2.free();
 	// draw the text
 	drawListBox(text, selection, nOptions, topIdx, da, fontDict, quadding,
-		    xMin, yMin, xMax, yMax, borderWidth);
+		    xMin, yMin, xMax, yMax, borderWidth, appearBuf);
 	for (i = 0; i < nOptions; ++i) {
 	  delete text[i];
 	}
@@ -979,7 +1322,22 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
       obj1.free();
     }
   } else if (ftObj.isName("Sig")) {
-    //~unimp
+    //~ check to see if background is already drawn
+    gfxStateDict.initDict(acroForm->doc->getXRef());
+    obj1.initReal(0.5);
+    gfxStateDict.dictAdd(copyString("ca"), &obj1);
+    appearBuf->append("/GS1 gs\n");
+    appearBuf->appendf("0.7 0.7 1 rg 0 0 {0:.2f} {1:.2f} re f\n",
+		       xMax - xMin, yMax - yMin);
+    caption = new GString("SIGN HERE");
+    if (da) {
+      delete da;
+    }
+    da = new GString("/Helv 10 Tf 1 0 0 rg");
+    drawText(caption, da, fontDict,
+	     gFalse, 0, acroFormQuadLeft, gFalse, gFalse, rot,
+	     xMin, yMin, xMax, yMax, borderWidth, appearBuf);
+    delete caption;
   } else {
     error(errSyntaxError, -1, "Unknown field type");
   }
@@ -1001,11 +1359,34 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
   obj1.arrayAdd(obj2.initReal(yMax - yMin));
   appearDict.dictAdd(copyString("BBox"), &obj1);
 
-  // set the resource dictionary
+  // set the resource dictionary; add a default font
   if (drObj.isDict()) {
-    appearDict.dictAdd(copyString("Resources"), drObj.copy(&obj1));
+    drObj.copy(&resources);
+  } else {
+    resources.initDict(acroForm->doc->getXRef());
   }
   drObj.free();
+  fontResources.initDict(acroForm->doc->getXRef());
+  if (resources.dictLookup("Font", &obj1)->isDict()) {
+    for (i = 0; i < obj1.dictGetLength(); ++i) {
+      obj1.dictGetValNF(i, &obj2);
+      fontResources.dictAdd(copyString(obj1.dictGetKey(i)), &obj2);
+    }
+  }
+  obj1.free();
+  defaultFont.initDict(acroForm->doc->getXRef());
+  defaultFont.dictAdd(copyString("Type"), obj1.initName("Font"));
+  defaultFont.dictAdd(copyString("Subtype"), obj1.initName("Type1"));
+  defaultFont.dictAdd(copyString("BaseFont"), obj1.initName("Helvetica"));
+  defaultFont.dictAdd(copyString("Encoding"), obj1.initName("WinAnsiEncoding"));
+  fontResources.dictAdd(copyString("xpdf_default_font"), &defaultFont);
+  resources.dictAdd(copyString("Font"), &fontResources);
+  if (gfxStateDict.isDict()) {
+    obj1.initDict(acroForm->doc->getXRef());
+    obj1.dictAdd(copyString("GS1"), &gfxStateDict);
+    resources.dictAdd(copyString("ExtGState"), &obj1);
+  }
+  appearDict.dictAdd(copyString("Resources"), &resources);
 
   // build the appearance stream
   appearStream = new MemStream(appearBuf->getCString(), 0,
@@ -1029,7 +1410,8 @@ void AcroFormField::drawNewAppearance(Gfx *gfx, Dict *annot,
 // have 1, 3, or 4 elements).  If <adjust> is +1, color is brightened;
 // if <adjust> is -1, color is darkened; otherwise color is not
 // modified.
-void AcroFormField::setColor(Array *a, GBool fill, int adjust) {
+void AcroFormField::setColor(Array *a, GBool fill, int adjust,
+			     GString *appearBuf) {
   Object obj1;
   double color[4];
   int nComps, i;
@@ -1078,7 +1460,7 @@ void AcroFormField::drawText(GString *text, GString *da, GfxFontDict *fontDict,
 			     GBool multiline, int comb, int quadding,
 			     GBool txField, GBool forceZapfDingbats, int rot,
 			     double xMin, double yMin, double xMax, double yMax,
-			     double border) {
+			     double border, GString *appearBuf) {
   GString *text2;
   GList *daToks;
   GString *tok;
@@ -1107,24 +1489,17 @@ void AcroFormField::drawText(GString *text, GString *da, GfxFontDict *fontDict,
   } else {
     text2 = text;
   }
+  if (text2->getLength() == 0) {
+    if (text2 != text) {
+      delete text2;
+    }
+    return;
+  }
 
   // parse the default appearance string
   tfPos = tmPos = -1;
   if (da) {
-    daToks = new GList();
-    i = 0;
-    while (i < da->getLength()) {
-      while (i < da->getLength() && Lexer::isSpace(da->getChar(i))) {
-	++i;
-      }
-      if (i < da->getLength()) {
-	for (j = i + 1;
-	     j < da->getLength() && !Lexer::isSpace(da->getChar(j));
-	     ++j) ;
-	daToks->append(new GString(da, i, j - i));
-	i = j;
-      }
-    }
+    daToks = tokenize(da);
     for (i = 2; i < daToks->getLength(); ++i) {
       if (i >= 2 && !((GString *)daToks->get(i))->cmp("Tf")) {
 	tfPos = i - 2;
@@ -1156,6 +1531,8 @@ void AcroFormField::drawText(GString *text, GString *da, GfxFontDict *fontDict,
     if (tok->getLength() >= 1 && tok->getChar(0) == '/') {
       if (!fontDict || !(font = fontDict->lookup(tok->getCString() + 1))) {
 	error(errSyntaxError, -1, "Unknown font in field's DA string");
+	tok->clear();
+	tok->append("/xpdf_default_font");
       }
     } else {
       error(errSyntaxError, -1,
@@ -1165,6 +1542,13 @@ void AcroFormField::drawText(GString *text, GString *da, GfxFontDict *fontDict,
     fontSize = atof(tok->getCString());
   } else {
     error(errSyntaxError, -1, "Missing 'Tf' operator in field's DA string");
+    fontSize = 10;
+    if (!daToks) {
+      daToks = new GList();
+    }
+    daToks->append(new GString("/xpdf_default_font"));
+    daToks->append(new GString("10"));
+    daToks->append(new GString("Tf"));
   }
 
   // setup
@@ -1226,7 +1610,11 @@ void AcroFormField::drawText(GString *text, GString *da, GfxFontDict *fontDict,
     // starting y coordinate
     // (note: each line of text starts with a Td operator that moves
     // down a line)
-    y = dy - 3;
+    if (dy > fontSize + 6) {
+      y = dy - 3;
+    } else {
+      y = 0.5 * dy - 0.4 * fontSize + fontSize;
+    }
 
     // set the font matrix
     if (tmPos >= 0) {
@@ -1278,11 +1666,11 @@ void AcroFormField::drawText(GString *text, GString *da, GfxFontDict *fontDict,
 	c = text2->getChar(i) & 0xff;
 	if (c == '(' || c == ')' || c == '\\') {
 	  appearBuf->append('\\');
-	  appearBuf->append(c);
+	  appearBuf->append((char)c);
 	} else if (c < 0x20 || c >= 0x80) {
 	  appearBuf->appendf("\\{0:03o}", c);
 	} else {
-	  appearBuf->append(c);
+	  appearBuf->append((char)c);
 	}
       }
       appearBuf->append(") Tj\n");
@@ -1364,11 +1752,11 @@ void AcroFormField::drawText(GString *text, GString *da, GfxFontDict *fontDict,
 	c = text2->getChar(i) & 0xff;
 	if (c == '(' || c == ')' || c == '\\') {
 	  appearBuf->append('\\');
-	  appearBuf->append(c);
+	  appearBuf->append((char)c);
 	} else if (c < 0x20 || c >= 0x80) {
 	  appearBuf->appendf("{0:.4f} 0 Td\n", w);
 	} else {
-	  appearBuf->append(c);
+	  appearBuf->append((char)c);
 	}
 	appearBuf->append(") Tj\n");
       }
@@ -1446,11 +1834,11 @@ void AcroFormField::drawText(GString *text, GString *da, GfxFontDict *fontDict,
 	c = text2->getChar(i) & 0xff;
 	if (c == '(' || c == ')' || c == '\\') {
 	  appearBuf->append('\\');
-	  appearBuf->append(c);
+	  appearBuf->append((char)c);
 	} else if (c < 0x20 || c >= 0x80) {
 	  appearBuf->appendf("\\{0:03o}", c);
 	} else {
-	  appearBuf->append(c);
+	  appearBuf->append((char)c);
 	}
       }
       appearBuf->append(") Tj\n");
@@ -1477,7 +1865,8 @@ void AcroFormField::drawListBox(GString **text, GBool *selection,
 				int nOptions, int topIdx,
 				GString *da, GfxFontDict *fontDict,
 				GBool quadding, double xMin, double yMin,
-				double xMax, double yMax, double border) {
+				double xMax, double yMax, double border,
+				GString *appearBuf) {
   GList *daToks;
   GString *tok;
   GfxFont *font;
@@ -1491,20 +1880,7 @@ void AcroFormField::drawListBox(GString **text, GBool *selection,
   // parse the default appearance string
   tfPos = tmPos = -1;
   if (da) {
-    daToks = new GList();
-    i = 0;
-    while (i < da->getLength()) {
-      while (i < da->getLength() && Lexer::isSpace(da->getChar(i))) {
-	++i;
-      }
-      if (i < da->getLength()) {
-	for (j = i + 1;
-	     j < da->getLength() && !Lexer::isSpace(da->getChar(j));
-	     ++j) ;
-	daToks->append(new GString(da, i, j - i));
-	i = j;
-      }
-    }
+    daToks = tokenize(da);
     for (i = 2; i < daToks->getLength(); ++i) {
       if (i >= 2 && !((GString *)daToks->get(i))->cmp("Tf")) {
 	tfPos = i - 2;
@@ -1524,6 +1900,8 @@ void AcroFormField::drawListBox(GString **text, GBool *selection,
     if (tok->getLength() >= 1 && tok->getChar(0) == '/') {
       if (!fontDict || !(font = fontDict->lookup(tok->getCString() + 1))) {
 	error(errSyntaxError, -1, "Unknown font in field's DA string");
+	tok->clear();
+	tok->append("/xpdf_default_font");
       }
     } else {
       error(errSyntaxError, -1,
@@ -1644,11 +2022,11 @@ void AcroFormField::drawListBox(GString **text, GBool *selection,
       c = text[i]->getChar(j) & 0xff;
       if (c == '(' || c == ')' || c == '\\') {
 	appearBuf->append('\\');
-	appearBuf->append(c);
+	appearBuf->append((char)c);
       } else if (c < 0x20 || c >= 0x80) {
 	appearBuf->appendf("\\{0:03o}", c);
       } else {
-	appearBuf->append(c);
+	appearBuf->append((char)c);
       }
     }
     appearBuf->append(") Tj\n");
@@ -1685,7 +2063,7 @@ void AcroFormField::getNextLine(GString *text, int start,
       break;
     }
     if (font && !font->isCIDFont()) {
-      dw = ((Gfx8BitFont *)font)->getWidth(c) * fontSize;
+      dw = ((Gfx8BitFont *)font)->getWidth((Guchar)c) * fontSize;
     } else {
       // otherwise, make a crude estimate
       dw = 0.5 * fontSize;
@@ -1735,7 +2113,7 @@ void AcroFormField::getNextLine(GString *text, int start,
 // Draw an (approximate) circle of radius <r> centered at (<cx>, <cy>).
 // <cmd> is used to draw the circle ("f", "s", or "b").
 void AcroFormField::drawCircle(double cx, double cy, double r,
-			       const char *cmd) {
+			       const char *cmd, GString *appearBuf) {
   appearBuf->appendf("{0:.4f} {1:.4f} m\n",
 		     cx + r, cy);
   appearBuf->appendf("{0:.4f} {1:.4f} {2:.4f} {3:.4f} {4:.4f} {5:.4f} c\n",
@@ -1759,7 +2137,8 @@ void AcroFormField::drawCircle(double cx, double cy, double r,
 
 // Draw the top-left half of an (approximate) circle of radius <r>
 // centered at (<cx>, <cy>).
-void AcroFormField::drawCircleTopLeft(double cx, double cy, double r) {
+void AcroFormField::drawCircleTopLeft(double cx, double cy, double r,
+				      GString *appearBuf) {
   double r2;
 
   r2 = r / sqrt(2.0);
@@ -1784,7 +2163,8 @@ void AcroFormField::drawCircleTopLeft(double cx, double cy, double r) {
 
 // Draw the bottom-right half of an (approximate) circle of radius <r>
 // centered at (<cx>, <cy>).
-void AcroFormField::drawCircleBottomRight(double cx, double cy, double r) {
+void AcroFormField::drawCircleBottomRight(double cx, double cy, double r,
+					  GString *appearBuf) {
   double r2;
 
   r2 = r / sqrt(2.0);
@@ -1807,6 +2187,27 @@ void AcroFormField::drawCircleBottomRight(double cx, double cy, double r) {
   appearBuf->append("S\n");
 }
 
+GList *AcroFormField::tokenize(GString *s) {
+  GList *toks;
+  int i, j;
+
+  toks = new GList();
+  i = 0;
+  while (i < s->getLength()) {
+    while (i < s->getLength() && Lexer::isSpace(s->getChar(i))) {
+      ++i;
+    }
+    if (i < s->getLength()) {
+      for (j = i + 1;
+	   j < s->getLength() && !Lexer::isSpace(s->getChar(j));
+	   ++j) ;
+      toks->append(new GString(s, i, j - i));
+      i = j;
+    }
+  }
+  return toks;
+}
+
 Object *AcroFormField::getResources(Object *res) {
   Object kidsObj, annotObj, obj1;
   int i;
@@ -1841,6 +2242,35 @@ Object *AcroFormField::getResources(Object *res) {
   return res;
 }
 
+Object *AcroFormField::getFieldRef(Object *ref) {
+  return fieldRef.copy(ref);
+}
+
+Object *AcroFormField::getValueObj(Object *val) {
+  return fieldLookup("V", val);
+}
+
+Object *AcroFormField::getParentRef(Object *parent) {
+  return fieldObj.dictLookupNF("Parent", parent);
+}
+
+// Get the first annotation object associated with this field.
+Object *AcroFormField::getAnnotObj(Object *annotObj) {
+  Object kidsObj;
+
+  if (fieldObj.dictLookup("Kids", &kidsObj)->isArray()) {
+    if (kidsObj.arrayGetLength() > 0) {
+      kidsObj.arrayGet(0, annotObj);
+    } else {
+      annotObj->initNull();
+    }
+  } else {
+    fieldObj.copy(annotObj);
+  }
+  kidsObj.free();
+  return annotObj;
+}
+
 Object *AcroFormField::getAnnotResources(Dict *annot, Object *res) {
   Object apObj, asObj, appearance, obj1;
 
@@ -1873,6 +2303,60 @@ Object *AcroFormField::getAnnotResources(Dict *annot, Object *res) {
   return res;
 }
 
+// Merge the field and AcroForm DR objects.
+void AcroFormField::buildDefaultResourceDict(Object *dr) {
+  Object formDR, fieldDR, resDict, newResDict, resObj;
+  char *resType, *resName;
+  int i, j;
+
+  // NB: we need to deep-copy the dictionaries here, because multiple
+  // threads can be sharing these objects.
+
+  dr->initDict(acroForm->doc->getXRef());
+
+  acroForm->acroFormObj.dictLookup("DR", &formDR);
+  if (formDR.isDict()) {
+    for (i = 0; i < formDR.dictGetLength(); ++i) {
+      resType = formDR.dictGetKey(i);
+      formDR.dictGetVal(i, &resDict);
+      if (resDict.isDict()) {
+	newResDict.initDict(acroForm->doc->getXRef());
+	dr->dictAdd(copyString(resType), &newResDict);
+	for (j = 0; j < resDict.dictGetLength(); ++j) {
+	  resName = resDict.dictGetKey(j);
+	  resDict.dictGetValNF(j, &resObj);
+	  newResDict.dictAdd(copyString(resName), &resObj);
+	}
+      }
+      resDict.free();
+    }
+  }
+  formDR.free();
+
+  fieldObj.dictLookup("DR", &fieldDR);
+  if (fieldDR.isDict()) {
+    for (i = 0; i < fieldDR.dictGetLength(); ++i) {
+      resType = fieldDR.dictGetKey(i);
+      fieldDR.dictGetVal(i, &resDict);
+      if (resDict.isDict()) {
+	dr->dictLookup(resType, &newResDict);
+	if (!newResDict.isDict()) {
+	  newResDict.free();
+	  newResDict.initDict(acroForm->doc->getXRef());
+	}
+	dr->dictAdd(copyString(resType), &newResDict);
+	for (j = 0; j < resDict.dictGetLength(); ++j) {
+	  resName = resDict.dictGetKey(j);
+	  resDict.dictGetValNF(j, &resObj);
+	  newResDict.dictAdd(copyString(resName), &resObj);
+	}
+      }
+      resDict.free();
+    }
+  }
+  fieldDR.free();
+}
+
 // Look up an inheritable field dictionary entry.
 Object *AcroFormField::fieldLookup(const char *key, Object *obj) {
   return fieldLookup(fieldObj.getDict(), key, obj);
